# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import multi_pex_api_pb2 as multi__pex__api__pb2


class MultiPexApiStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreatePexServer = channel.unary_unary(
                '/multi_pex_api.MultiPexApi/CreatePexServer',
                request_serializer=multi__pex__api__pb2.CreatePexServerRequest.SerializeToString,
                response_deserializer=multi__pex__api__pb2.CreatePexServerReply.FromString,
                )
        self.GetPexServers = channel.unary_unary(
                '/multi_pex_api.MultiPexApi/GetPexServers',
                request_serializer=multi__pex__api__pb2.GetPexServersRequest.SerializeToString,
                response_deserializer=multi__pex__api__pb2.GetPexServersReply.FromString,
                )
        self.GetCrashedPexServers = channel.unary_unary(
                '/multi_pex_api.MultiPexApi/GetCrashedPexServers',
                request_serializer=multi__pex__api__pb2.GetCrashedPexServersRequest.SerializeToString,
                response_deserializer=multi__pex__api__pb2.GetCrashedPexServersReply.FromString,
                )
        self.ShutdownPexServer = channel.unary_unary(
                '/multi_pex_api.MultiPexApi/ShutdownPexServer',
                request_serializer=multi__pex__api__pb2.ShutdownPexServerRequest.SerializeToString,
                response_deserializer=multi__pex__api__pb2.ShutdownPexServerReply.FromString,
                )
        self.Ping = channel.unary_unary(
                '/multi_pex_api.MultiPexApi/Ping',
                request_serializer=multi__pex__api__pb2.PingRequest.SerializeToString,
                response_deserializer=multi__pex__api__pb2.PingReply.FromString,
                )


class MultiPexApiServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreatePexServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPexServers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCrashedPexServers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShutdownPexServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MultiPexApiServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreatePexServer': grpc.unary_unary_rpc_method_handler(
                    servicer.CreatePexServer,
                    request_deserializer=multi__pex__api__pb2.CreatePexServerRequest.FromString,
                    response_serializer=multi__pex__api__pb2.CreatePexServerReply.SerializeToString,
            ),
            'GetPexServers': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPexServers,
                    request_deserializer=multi__pex__api__pb2.GetPexServersRequest.FromString,
                    response_serializer=multi__pex__api__pb2.GetPexServersReply.SerializeToString,
            ),
            'GetCrashedPexServers': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCrashedPexServers,
                    request_deserializer=multi__pex__api__pb2.GetCrashedPexServersRequest.FromString,
                    response_serializer=multi__pex__api__pb2.GetCrashedPexServersReply.SerializeToString,
            ),
            'ShutdownPexServer': grpc.unary_unary_rpc_method_handler(
                    servicer.ShutdownPexServer,
                    request_deserializer=multi__pex__api__pb2.ShutdownPexServerRequest.FromString,
                    response_serializer=multi__pex__api__pb2.ShutdownPexServerReply.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=multi__pex__api__pb2.PingRequest.FromString,
                    response_serializer=multi__pex__api__pb2.PingReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'multi_pex_api.MultiPexApi', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MultiPexApi(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreatePexServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/multi_pex_api.MultiPexApi/CreatePexServer',
            multi__pex__api__pb2.CreatePexServerRequest.SerializeToString,
            multi__pex__api__pb2.CreatePexServerReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetPexServers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/multi_pex_api.MultiPexApi/GetPexServers',
            multi__pex__api__pb2.GetPexServersRequest.SerializeToString,
            multi__pex__api__pb2.GetPexServersReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetCrashedPexServers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/multi_pex_api.MultiPexApi/GetCrashedPexServers',
            multi__pex__api__pb2.GetCrashedPexServersRequest.SerializeToString,
            multi__pex__api__pb2.GetCrashedPexServersReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ShutdownPexServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/multi_pex_api.MultiPexApi/ShutdownPexServer',
            multi__pex__api__pb2.ShutdownPexServerRequest.SerializeToString,
            multi__pex__api__pb2.ShutdownPexServerReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/multi_pex_api.MultiPexApi/Ping',
            multi__pex__api__pb2.PingRequest.SerializeToString,
            multi__pex__api__pb2.PingReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from dagster import Field, Permissive


def meter_provider_config_schema() -> Field:
    return Field(
        config=Permissive(
            description="Provide configuration for an OpenTelemetry meter provider.", fields={}
        )
    )



from .defs import (
    build_anomaly_detection_freshness_checks as build_anomaly_detection_freshness_checks,
)
from .types import (
    AnomalyDetectionModelParams as AnomalyDetectionModelParams,
    BetaFreshnessAnomalyDetectionParams as BetaFreshnessAnomalyDetectionParams,
)



from typing import NamedTuple, Optional

from dagster._serdes import whitelist_for_serdes
from dagster_cloud_cli.core.agent_queue import AgentQueue


@whitelist_for_serdes
class AgentQueuesConfig(NamedTuple):
    include_default_queue: bool = True
    additional_queues: Optional[list[AgentQueue]] = None

    @staticmethod
    def default_queues() -> list[AgentQueue]:
        return [None]

    @property
    def queues(self) -> list[AgentQueue]:
        if self.additional_queues:
            queues = self.additional_queues.copy()
        else:
            queues = []
        if self.include_default_queue:
            for queue in AgentQueuesConfig.default_queues():
                queues.append(queue)
        return queues

    def matches(self, queue: AgentQueue):
        return queue in self.queues



from dagster import BoolSource, Enum, Field, IntSource, Map, Permissive, Shape, StringSource

from ..enum import (
    AggregationTemporalityEnum,
    CompressionEnum,
    LoggingExporterEnum,
    MetricsExporterEnum,
    MetricsInstrumentTypesEnum,
    TracingExporterEnum,
    ViewAggregationEnum,
)

# Convert Python to Dagster enums
compression_enum = Enum.from_python_enum_direct_values(CompressionEnum)
logging_exporter_enum = Enum.from_python_enum_direct_values(LoggingExporterEnum)
metrics_exporter_enum = Enum.from_python_enum_direct_values(MetricsExporterEnum)
tracing_exporter_enum = Enum.from_python_enum_direct_values(TracingExporterEnum)

aggregation_temporality_enum = Enum.from_python_enum_direct_values(AggregationTemporalityEnum)
metrics_instrument_types_enum = Enum.from_python_enum_direct_values(MetricsInstrumentTypesEnum)
view_aggregation_enum = Enum.from_python_enum_direct_values(ViewAggregationEnum)


def exporter_config_schema(pillar: str) -> Shape:
    """Different observability pillars have different supported exporters but share common parameters.
    Valid pillars are: logging, metrics, tracing.
    """
    exporter_type_field: Field
    exporter_extra_params: dict = {}
    if pillar == "logging":
        exporter_type_field = Field(
            config=logging_exporter_enum,
            is_required=False,
            default_value=LoggingExporterEnum.ConsoleLogExporter.value,
        )
    elif pillar == "metrics":
        exporter_type_field = Field(
            config=metrics_exporter_enum,
            is_required=False,
            default_value=MetricsExporterEnum.ConsoleMetricExporter.value,
        )
        exporter_extra_params = {
            "preferred_temporality": Field(
                config=Map(str, aggregation_temporality_enum), is_required=False
            ),
            "preferred_aggregation": Field(
                config=Map(str, view_aggregation_enum), is_required=False
            ),
        }
    elif pillar == "tracing":
        exporter_type_field = Field(
            config=tracing_exporter_enum,
        )
    else:
        raise ValueError(f"Invalid pillar: {pillar}")

    return Shape(
        description="Provide configuration for an OpenTelemetry exporter.",
        fields={
            "type": exporter_type_field,
            "params": Permissive(
                fields={
                    # These fields are common to most exporters. No defaults are provided, to allow either
                    # OpenTelemetry's default or environment variables to be used.
                    # Ref: https://opentelemetry-python.readthedocs.io/en/latest/sdk/environment_variables.html
                    "endpoint": Field(StringSource, is_required=False),
                    "compression": Field(config=compression_enum, is_required=False),
                    "headers": Field(
                        description="Custom headers to send", config=Permissive(), is_required=False
                    ),
                    "certificate_file": Field(StringSource, is_required=False),
                    "client_key_file": Field(StringSource, is_required=False),
                    "client_certificate_file": Field(StringSource, is_required=False),
                    "insecure": Field(BoolSource, is_required=False),
                    "timeout": Field(IntSource, is_required=False),
                    # These fields are specific an exporter.
                    **exporter_extra_params,
                }
            ),
        },
    )



from abc import abstractproperty
from enum import Enum
from typing import Any

from dagster import DagsterError
from pydantic import BaseModel


class AnomalyDetectionError(DagsterError):
    pass


class FreshnessAnomalyDetectionResult(BaseModel):
    last_updated_timestamp: float
    evaluation_timestamp: float
    last_update_time_lower_bound: float
    maximum_acceptable_delta: float
    model_training_range_start_timestamp: float
    model_training_range_end_timestamp: float


class AnomalyDetectionModelVersion(Enum):
    FRESHNESS_BETA = "FRESHNESS_BETA"

    @property
    def minimum_required_records(self) -> int:
        if self == AnomalyDetectionModelVersion.FRESHNESS_BETA:
            return 15
        raise NotImplementedError(f"Minimum required records not implemented for {self}")

    @property
    def minimum_required_records_msg(self) -> str:
        return f"Not enough records found to detect anomalies. Need at least {self.minimum_required_records}."


### INTERNAL MODEL PARAMETER SETS ###


class InternalModelParams(BaseModel):
    @abstractproperty
    def model_version(self) -> AnomalyDetectionModelVersion:
        raise NotImplementedError("Subclasses must implement this method")


class InternalBetaFreshnessAnomalyDetectionParams(InternalModelParams):
    sensitivity: float
    asset_key_user_string: str

    @property
    def model_version(self) -> AnomalyDetectionModelVersion:
        return AnomalyDetectionModelVersion.FRESHNESS_BETA


### USER FACING MODEL PARAMETER SETS ###


class AnomalyDetectionModelParams(BaseModel):
    @abstractproperty
    def model_version(self) -> AnomalyDetectionModelVersion:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractproperty
    def as_metadata(self) -> dict[str, Any]:
        raise NotImplementedError("Subclasses must implement this method")


class BetaFreshnessAnomalyDetectionParams(AnomalyDetectionModelParams):
    sensitivity: float

    @property
    def model_version(self) -> AnomalyDetectionModelVersion:
        return AnomalyDetectionModelVersion.FRESHNESS_BETA

    @property
    def as_metadata(self) -> dict[str, Any]:
        return {
            "sensitivity": self.sensitivity,
        }



import re

from ..user_code_launcher.utils import unique_resource_name


def unique_docker_resource_name(deployment_name, location_name):
    return unique_resource_name(
        deployment_name,
        location_name,
        length_limit=63,  # Max DNS hostname limit
        sanitize_fn=lambda name: re.sub("[^a-z0-9-]", "", name).strip("-"),
    )



import enum
from collections.abc import Mapping, Sequence
from datetime import timedelta
from enum import Enum
from typing import Any, Optional, TypedDict, Union

from dagster._core.code_pointer import CodePointer
from dagster._core.definitions.selector import JobSelector
from dagster._core.events.log import EventLogEntry
from dagster._core.remote_representation import CodeLocationOrigin, RepositorySnap
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.utils import RequestUtilizationMetrics
from dagster._record import IHaveNew, copy, record, record_custom
from dagster._serdes import whitelist_for_serdes
from dagster._time import get_current_datetime, get_current_timestamp
from dagster._utils.container import ContainerUtilizationMetrics
from dagster._utils.error import SerializableErrorInfo
from dagster_cloud_cli.core.workspace import CodeLocationDeployData
from typing_extensions import NotRequired

from dagster_cloud.agent import AgentQueuesConfig
from dagster_cloud.execution.monitoring import (
    CloudCodeServerHeartbeat,
    CloudContainerResourceLimits,
    CloudRunWorkerStatuses,
)
from dagster_cloud.util import keys_not_none

DEFAULT_EXPIRATION_MILLISECONDS = 10 * 60 * 1000


@whitelist_for_serdes
@record
class DagsterCloudUploadRepositoryData:
    """Serialized object uploaded by the Dagster Cloud agent with information pulled
    from a gRPC server about an individual repository - the data field is serialized since the
    agent may be running older code that doesn't know how to deserialize it, so it passes
    it serialized up to the host cloud, which is always up to date.
    """

    repository_name: str
    code_pointer: CodePointer
    serialized_repository_data: str


@whitelist_for_serdes
@record
class DagsterCloudUploadLocationData:
    """Serialized object uploaded by the Dagster Cloud agent with information pulled
    about a successfully loaded repository location, including information about
    each repository as well as shared metadata like the image to use when launching
    runs in this location.
    """

    upload_repository_datas: list[DagsterCloudUploadRepositoryData]
    container_image: Optional[str]
    executable_path: Optional[str]
    dagster_library_versions: Optional[Mapping[str, str]] = None


@whitelist_for_serdes(storage_field_names={"code_location_deploy_data": "deployment_metadata"})
@record
class DagsterCloudUploadWorkspaceEntry:
    """Serialized object uploaded by the Dagster Cloud agent with information about
    a repository location - either the serialized DagsterCloudUploadLocationData
    if the location loaded succesfully, or a SerializableErrorInfo describing the
    error if it was not.
    """

    location_name: str
    code_location_deploy_data: CodeLocationDeployData
    upload_location_data: Optional[DagsterCloudUploadLocationData]
    serialized_error_info: Optional[SerializableErrorInfo]


@whitelist_for_serdes
@record
class DagsterCloudUploadWorkspaceResponse:
    updated: bool
    message: str
    missing_job_snapshots: Optional[Sequence[JobSelector]]


@whitelist_for_serdes
class DagsterCloudApi(Enum):
    CHECK_FOR_WORKSPACE_UPDATES = "CHECK_FOR_WORKSPACE_UPDATES"
    GET_EXTERNAL_EXECUTION_PLAN = "GET_EXTERNAL_EXECUTION_PLAN"
    GET_SUBSET_EXTERNAL_PIPELINE_RESULT = "GET_SUBSET_EXTERNAL_PIPELINE_RESULT"
    GET_EXTERNAL_PARTITION_CONFIG = "GET_EXTERNAL_PARTITION_CONFIG"
    GET_EXTERNAL_PARTITION_TAGS = "GET_EXTERNAL_PARTITION_TAGS"
    GET_EXTERNAL_PARTITION_NAMES = "GET_EXTERNAL_PARTITION_NAMES"
    GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA = (
        "GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA"
    )
    GET_EXTERNAL_SCHEDULE_EXECUTION_DATA = "GET_EXTERNAL_SCHEDULE_EXECUTION_DATA"
    GET_EXTERNAL_SENSOR_EXECUTION_DATA = "GET_EXTERNAL_SENSOR_EXECUTION_DATA"
    GET_EXTERNAL_NOTEBOOK_DATA = "GET_EXTERNAL_NOTEBOOK_DATA"

    LAUNCH_RUN = "LAUNCH_RUN"
    TERMINATE_RUN = "TERMINATE_RUN"

    PING_LOCATION = "PING_LOCATION"  # Signal that a location is in use and should keep servers up

    CHECK_RUN_HEALTH = "CHECK_RUN_HEALTH"  # deprecated, agents now surface this in heartbeats
    LAUNCH_STEP = "LAUNCH_STEP"  # deprecated with cloud executor
    CHECK_STEP_HEALTH = "CHECK_STEP_HEALTH"  # deprecated with cloud executor
    TERMINATE_STEP = "TERMINATE_STEP"  # deprecated with cloud executor
    LOAD_REPOSITORIES = "LOAD_REPOSITORIES"  # deprecated

    def __structlog__(self):
        return self.name


@whitelist_for_serdes
@record
class DagsterCloudApiThreadTelemetry:
    submitted_to_executor_timestamp: float
    thread_start_run_timestamp: float
    thread_end_handle_api_request_timestamp: float

    @property
    def time_to_thread_initialization_seconds(self) -> float:
        return self.thread_start_run_timestamp - self.submitted_to_executor_timestamp

    @property
    def time_to_handle_api_request_seconds(self) -> float:
        return self.thread_end_handle_api_request_timestamp - self.thread_start_run_timestamp


@whitelist_for_serdes(old_storage_names={"CheckForCodeLocationUpdatesRequest"})
@record
class DagsterCloudApiSuccess:
    thread_telemetry: Optional[DagsterCloudApiThreadTelemetry] = None

    def with_thread_telemetry(self, thread_telemetry: DagsterCloudApiThreadTelemetry):
        return copy(self, thread_telemetry=thread_telemetry)


@whitelist_for_serdes
@record
class DagsterCloudApiUnknownCommandResponse:
    request_api: str
    thread_telemetry: Optional[DagsterCloudApiThreadTelemetry] = None

    def with_thread_telemetry(self, thread_telemetry: DagsterCloudApiThreadTelemetry):
        return copy(self, thread_telemetry=thread_telemetry)


@whitelist_for_serdes
@record
class DagsterCloudApiErrorResponse:
    error_infos: list[SerializableErrorInfo]
    thread_telemetry: Optional[DagsterCloudApiThreadTelemetry] = None

    def with_thread_telemetry(self, thread_telemetry: DagsterCloudApiThreadTelemetry):
        return copy(self, thread_telemetry=thread_telemetry)


@whitelist_for_serdes
@record
class DagsterCloudApiGrpcResponse:
    # Class that DagsterCloudApi methods can use to pass along the result of
    # a gRPC call against the user code server. The field here is passed in
    # serialized as a string, because we can't guarantee that the agent code will
    # be up-to-date enough to know how to deserialize it (but the host cloud always
    # should, since it will always be up to date).
    serialized_response_or_error: str
    thread_telemetry: Optional[DagsterCloudApiThreadTelemetry] = None

    def with_thread_telemetry(self, thread_telemetry: DagsterCloudApiThreadTelemetry):
        return copy(self, thread_telemetry=thread_telemetry)


@whitelist_for_serdes
@record
class LoadRepositoriesArgs:
    location_origin: CodeLocationOrigin


@whitelist_for_serdes
@record
class DagsterCloudRepositoryData:
    repo_name: str
    code_pointer: CodePointer
    external_repository_data: RepositorySnap


@whitelist_for_serdes(storage_field_names={"code_location_deploy_data": "code_deployment_metadata"})
@record
class LoadRepositoriesResponse:
    repository_datas: Sequence[DagsterCloudRepositoryData]
    container_image: Optional[str]
    executable_path: Optional[str]
    code_location_deploy_data: Optional[CodeLocationDeployData] = None
    dagster_library_versions: Optional[Mapping[str, str]] = None


@whitelist_for_serdes
@record
class PingLocationArgs:
    location_name: str


@whitelist_for_serdes(storage_field_names={"dagster_run": "pipeline_run"})
@record
class LaunchRunArgs:
    dagster_run: DagsterRun


@whitelist_for_serdes(storage_field_names={"dagster_run": "pipeline_run"})
@record
class TerminateRunArgs:
    dagster_run: DagsterRun


@whitelist_for_serdes
@record_custom
class DagsterCloudApiRequest(IHaveNew):
    request_id: str
    request_api: DagsterCloudApi
    request_args: Any
    deployment_name: str
    expire_at: float
    is_branch_deployment: bool

    def __new__(
        cls,
        request_id: str,
        request_api: DagsterCloudApi,
        request_args: Any,
        deployment_name: str,
        expire_at: Optional[float] = None,
        is_branch_deployment: Optional[bool] = None,
    ):
        return super().__new__(
            cls,
            request_id=request_id,
            request_api=request_api,
            request_args=request_args,
            deployment_name=deployment_name,
            expire_at=expire_at
            if expire_at is not None
            else (
                get_current_datetime() + timedelta(milliseconds=DEFAULT_EXPIRATION_MILLISECONDS)
            ).timestamp(),
            is_branch_deployment=False if is_branch_deployment is None else is_branch_deployment,
        )

    @property
    def is_expired(self) -> bool:
        return get_current_timestamp() > self.expire_at

    @staticmethod
    def format_request(request_id: str, request_api: Union[str, DagsterCloudApi]) -> str:
        return f"[{request_id}: {request_api}]"

    def __str__(self) -> str:
        return DagsterCloudApiRequest.format_request(self.request_id, self.request_api)


DagsterCloudApiResponse = Union[
    DagsterCloudApiSuccess,
    DagsterCloudApiGrpcResponse,
    DagsterCloudApiErrorResponse,
    DagsterCloudApiUnknownCommandResponse,
]

DagsterCloudApiResponseTypesTuple = (
    DagsterCloudApiSuccess,
    DagsterCloudApiGrpcResponse,
    DagsterCloudApiErrorResponse,
    DagsterCloudApiUnknownCommandResponse,
)


@whitelist_for_serdes
@record
class StoreEventBatchRequest:
    event_log_entries: Sequence[EventLogEntry]


@whitelist_for_serdes
@record
class DagsterCloudUploadApiResponse:
    request_id: str
    request_api: str
    response: DagsterCloudApiResponse


@whitelist_for_serdes
@record
class BatchDagsterCloudUploadApiResponse:
    batch: list[DagsterCloudUploadApiResponse]


@whitelist_for_serdes
@record
class TimestampedError:
    timestamp: Optional[float]
    error: SerializableErrorInfo


class UserCodeDeploymentType(enum.Enum):
    SERVERLESS = "serverless"
    ECS = "ecs"
    K8S = "k8s"
    DOCKER = "docker"
    PROCESS = "process"
    UNKNOWN = "unknown"

    @property
    def supports_utilization_metrics(self) -> bool:
        return self in [
            UserCodeDeploymentType.ECS,
            UserCodeDeploymentType.K8S,
            UserCodeDeploymentType.SERVERLESS,
        ]


class AgentUtilizationMetrics(TypedDict):
    container_utilization: ContainerUtilizationMetrics
    request_utilization: RequestUtilizationMetrics
    resource_limits: CloudContainerResourceLimits


class AgentHeartbeatMetadata(TypedDict):
    utilization_metrics: NotRequired[AgentUtilizationMetrics]
    version: NotRequired[str]
    image_tag: NotRequired[str]
    type: NotRequired[str]
    queues: NotRequired[list[str]]


@whitelist_for_serdes
@record_custom
class AgentHeartbeat(IHaveNew):
    timestamp: float
    agent_id: str
    agent_label: Optional[str]
    agent_type: Optional[str]
    errors: Optional[Sequence[TimestampedError]]
    metadata: AgentHeartbeatMetadata
    run_worker_statuses: Optional[CloudRunWorkerStatuses]
    code_server_heartbeats: Sequence[CloudCodeServerHeartbeat]
    agent_queues_config: AgentQueuesConfig

    def __new__(
        cls,
        timestamp: float,
        agent_id: str,
        agent_label: Optional[str],
        agent_type: Optional[str],
        errors: Optional[Sequence[TimestampedError]] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        run_worker_statuses: Optional[CloudRunWorkerStatuses] = None,
        code_server_heartbeats: Optional[Sequence[CloudCodeServerHeartbeat]] = None,
        agent_queues_config: Optional[AgentQueuesConfig] = None,
    ):
        return super().__new__(
            cls,
            timestamp=timestamp,
            agent_id=agent_id,
            agent_label=agent_label,
            agent_type=agent_type,
            errors=errors,
            metadata=metadata or {},
            run_worker_statuses=run_worker_statuses,
            code_server_heartbeats=code_server_heartbeats or [],
            agent_queues_config=agent_queues_config or AgentQueuesConfig(),
        )

    def without_messages_and_errors(self) -> "AgentHeartbeat":
        return copy(
            self,
            errors=[],
            run_worker_statuses=self.run_worker_statuses.without_messages_and_errors()
            if self.run_worker_statuses
            else None,
            code_server_heartbeats=[
                heartbeat.without_messages_and_errors() for heartbeat in self.code_server_heartbeats
            ],
            metadata={
                key: val for key, val in self.metadata.items() if key != "utilization_metrics"
            },
        )

    def get_agent_utilization_metrics(self) -> Optional[AgentUtilizationMetrics]:
        metrics = self.metadata.get("utilization_metrics")
        if metrics and keys_not_none(["container_utilization", "request_utilization"], metrics):
            return metrics
        return None


class FileFormat:
    JSON = "json"
    GZIPPED_JSON = "json.gz"


class SnapshotType:
    ERROR = "error"
    JOB = "job"
    REPOSITORY = "repository"


@whitelist_for_serdes
@record
class StoredSnapshot:
    sha1: str
    format: str
    uri: str
    decompressed_size: int


@whitelist_for_serdes
@record
class SnapshotUploadData:
    sha1: str
    format: str
    uri: str
    presigned_put_url: str
    id: str
    type: str


@whitelist_for_serdes
@record
class CheckSnapshotResult:
    stored_snapshot: Optional[StoredSnapshot]
    upload_data: Optional[SnapshotUploadData]


@whitelist_for_serdes
@record
class ConfirmUploadResult:
    stored_snapshot: StoredSnapshot


@whitelist_for_serdes
@record
class DagsterCloudRepositoryManifest:
    name: str
    code_pointer: CodePointer
    stored_snapshot: StoredSnapshot


@whitelist_for_serdes
@record
class DagsterCloudCodeLocationManifest:
    repositories: Sequence[DagsterCloudRepositoryManifest]
    container_image: Optional[str]
    executable_path: Optional[str]
    dagster_library_versions: Optional[Mapping[str, str]]
    code_location_deploy_data: CodeLocationDeployData


@whitelist_for_serdes
@record
class DagsterCloudCodeLocationUpdateResult:
    location_name: str
    manifest: Optional[DagsterCloudCodeLocationManifest]
    error_snapshot: Optional[StoredSnapshot]


@whitelist_for_serdes
@record
class DagsterCloudCodeLocationUpdateResponse:
    updated: bool
    message: str
    missing_job_snapshots: Optional[Sequence[JobSelector]]



from .client import (
    MultiPexGrpcClient as MultiPexGrpcClient,
    wait_for_grpc_server as wait_for_grpc_server,
)
from .types import (
    CreatePexServerArgs as CreatePexServerArgs,
    GetPexServersArgs as GetPexServersArgs,
    PexServerHandle as PexServerHandle,
    ShutdownPexServerArgs as ShutdownPexServerArgs,
)



"""Compile the proto definitions into Python.

This tooling should be invoked to regenerate the Python grpc artifacts by running:

    python -m dagster_cloud.pex.grpc.compile
"""

import os
import shutil
import subprocess
import sys

import pkg_resources
from dagster._utils import file_relative_path, safe_tempfile_path

PROTOS_DIR = file_relative_path(__file__, "protos")

PROTOS_PATH = os.path.join(PROTOS_DIR, "multi_pex_api.proto")

GENERATED_DIR = file_relative_path(__file__, "__generated__")

ISORT_SETTINGS_PATH = file_relative_path(__file__, "../../../../")

GENERATED_HEADER = [
    (
        "# @" + "generated\n"  # This is to avoid matching the phab rule
    ),
    "\n",
    "# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`\n",
    "# Do not edit this file directly, and do not attempt to recompile it using\n",
    "# grpc_tools.protoc directly, as several changes must be made to the raw output\n",
    "\n",
]

GENERATED_PB2_PYLINT_DIRECTIVE = [
    "# ruff: noqa: SLF001\n",
    "\n",
]


def protoc(generated_dir: str):
    generated_pb2_path = os.path.join(generated_dir, "multi_pex_api_pb2.py")
    generated_grpc_path = os.path.join(generated_dir, "multi_pex_api_pb2_grpc.py")

    # python -m grpc_tools.protoc \
    #   -I protos --python_out __generated__ --grpc_python_out __generated__ protos/multi_pex_api.proto
    _res = subprocess.check_output(
        [
            sys.executable,
            "-m",
            "grpc_tools.protoc",
            "-I",
            PROTOS_DIR,
            "--python_out",
            generated_dir,
            "--grpc_python_out",
            generated_dir,
            "--mypy_out",
            generated_dir,
            PROTOS_PATH,
        ]
    )

    # The generated multi_pex_api_pb2_grpc.py file must be altered:
    # 1. Change the import from `import multi_pex_api_pb2 as multi_pex_api__pb2` to `from . import multi_pex_api_pb2 as multi_pex_api__pb2`.
    #    See: https://github.com/grpc/grpc/issues/22914
    with safe_tempfile_path() as tempfile_path:
        shutil.copyfile(
            generated_grpc_path,
            tempfile_path,
        )
        with open(tempfile_path, encoding="utf8") as generated:
            with open(generated_grpc_path, "w", encoding="utf8") as rewritten:
                for line in GENERATED_HEADER:
                    rewritten.write(line)

                for line in generated.readlines():
                    if line == "import multi_pex_api_pb2 as multi__pex__api__pb2\n":
                        rewritten.write("from . import multi_pex_api_pb2 as multi__pex__api__pb2\n")
                    else:
                        rewritten.write(line)

    with safe_tempfile_path() as tempfile_path:
        shutil.copyfile(
            generated_pb2_path,
            tempfile_path,
        )
        with open(tempfile_path, encoding="utf8") as generated:
            with open(generated_pb2_path, "w", encoding="utf8") as rewritten:
                for line in GENERATED_HEADER:
                    rewritten.write(line)

                for line in GENERATED_PB2_PYLINT_DIRECTIVE:
                    rewritten.write(line)

                for line in generated.readlines():
                    rewritten.write(line)

    installed_pkgs = {pkg.key for pkg in pkg_resources.working_set}

    # Run ruff if it's available. This is under a conditional because ruff may not be available in
    # a test environment.
    if "ruff" in installed_pkgs:
        subprocess.check_output(
            [
                "ruff",
                "format",
                "--line-length=100",
                generated_dir,
            ]
        )


if __name__ == "__main__":
    protoc(sys.argv[1] if len(sys.argv) > 1 else GENERATED_DIR)



from collections.abc import Generator
from contextlib import AbstractContextManager, contextmanager
from typing import Optional, Protocol


class Instrumentation(Protocol):
    def tags(self, tags: list[str]) -> "Instrumentation": ...

    def histogram(self, name: str, value: float) -> None: ...

    def increment(self, name: str) -> None: ...

    def instrument_context(
        self, name: str, buckets_ms: Optional[list[int]]
    ) -> AbstractContextManager[None]: ...


class NoOpInstrumentation(Instrumentation):
    def tags(self, tags: list[str]) -> Instrumentation:
        return self

    def histogram(self, name: str, value: float) -> None:
        pass

    def increment(self, name: str) -> None:
        pass

    @contextmanager
    def instrument_context(
        self, name: str, buckets_ms: Optional[list[int]]
    ) -> Generator[None, None, None]:
        yield



from collections.abc import Sequence
from typing import Optional

from dagster._serdes import serialize_value
from dagster._utils.error import SerializableErrorInfo

ERROR_CLASS_NAME_SIZE_LIMIT = 1000


def unwrap_user_code_error(error_info: SerializableErrorInfo) -> SerializableErrorInfo:
    """Extracts the underlying error from the passed error, if it is a DagsterUserCodeLoadError."""
    if error_info.cls_name == "DagsterUserCodeLoadError":
        return unwrap_user_code_error(error_info.cause)
    return error_info


def truncate_serialized_error(
    error_info: SerializableErrorInfo,
    field_size_limit: int,
    max_depth: int,
    truncations: Optional[list[str]] = None,
):
    truncations = [] if truncations is None else truncations

    if error_info.cause:
        if max_depth == 0:
            truncations.append("cause")
            new_cause = (
                error_info.cause
                if len(serialize_value(error_info.cause)) <= field_size_limit
                else SerializableErrorInfo(
                    message="(Cause truncated due to size limitations)",
                    stack=[],
                    cls_name=None,
                )
            )
        else:
            new_cause = truncate_serialized_error(
                error_info.cause,
                field_size_limit,
                max_depth=max_depth - 1,
                truncations=truncations,
            )
        error_info = error_info._replace(cause=new_cause)

    if error_info.context:
        if max_depth == 0:
            truncations.append("context")
            new_context = (
                error_info.context
                if len(serialize_value(error_info.context)) <= field_size_limit
                else SerializableErrorInfo(
                    message="(Context truncated due to size limitations)",
                    stack=[],
                    cls_name=None,
                )
            )
        else:
            new_context = truncate_serialized_error(
                error_info.context,
                field_size_limit,
                max_depth=max_depth - 1,
                truncations=truncations,
            )
        error_info = error_info._replace(context=new_context)

    stack_size_so_far = 0
    truncated_stack = []
    for stack_elem in error_info.stack:
        stack_size_so_far += len(stack_elem)
        if stack_size_so_far > field_size_limit:
            truncations.append("stack")
            truncated_stack.append("(TRUNCATED)")
            break

        truncated_stack.append(stack_elem)

    error_info = error_info._replace(stack=truncated_stack)

    msg_len = len(error_info.message)
    if msg_len > field_size_limit:
        truncations.append(f"message from {msg_len} to {field_size_limit}")
        error_info = error_info._replace(
            message=error_info.message[:field_size_limit] + " (TRUNCATED)"
        )

    if error_info.cls_name and len(error_info.cls_name) > ERROR_CLASS_NAME_SIZE_LIMIT:
        truncations.append("cls_name")
        error_info = error_info._replace(
            cls_name=error_info.cls_name[:ERROR_CLASS_NAME_SIZE_LIMIT] + " (TRUNCATED)"
        )

    return error_info


DAGSTER_FRAMEWORK_SUBSTRINGS = [
    "/site-packages/dagster/",
    "/python_modules/dagster/dagster",
]


def remove_dagster_framework_lines_from_serializable_exc_info(error_info: SerializableErrorInfo):
    return error_info._replace(
        stack=remove_dagster_framework_lines_from_stack_trace(error_info.stack),
        cause=(
            remove_dagster_framework_lines_from_serializable_exc_info(error_info.cause)
            if error_info.cause
            else None
        ),
        context=(
            remove_dagster_framework_lines_from_serializable_exc_info(error_info.context)
            if error_info.context
            else None
        ),
    )


def remove_dagster_framework_lines_from_stack_trace(
    stack: Sequence[str],
) -> Sequence[str]:
    # Remove lines until you find the first non-dagster framework line

    for i in range(len(stack)):
        if not _line_contains_dagster_framework_file(stack[i]):
            return stack[i:]

    # Return the full stack trace if its all Dagster framework lines,
    # to not be left with an empty stack trace
    return stack


def _line_contains_dagster_framework_file(line: str):
    # stack trace line starts with something like
    # File "/usr/local/lib/python3.11/site-packages/dagster/_core/execution/plan/utils.py",
    split_by_comma = line.split(",")
    if not split_by_comma:
        return False

    file_portion = split_by_comma[0]
    return any(
        framework_substring in file_portion for framework_substring in DAGSTER_FRAMEWORK_SUBSTRINGS
    )



from typing import Optional

from opentelemetry._logs import LoggerProvider as APILoggerProvider
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler, LogRecordProcessor
from opentelemetry.sdk._logs.export import LogExporter
from opentelemetry.sdk.resources import Resource

from dagster_cloud.opentelemetry.enum import LoggingExporterEnum, LogRecordProcessorEnum


def build_log_exporter(exporter_config: dict) -> LogExporter:
    """Build an OpenTelemetry log exporter.
    params:
        exporter_config: dict: The exporter configuration.

    Returns:
        LogExporter: The log exporter instance.
    """
    exporter_type = exporter_config.get("type", LoggingExporterEnum.ConsoleLogExporter.value)
    exporter_params = exporter_config.get("params", {})

    if exporter_type == LoggingExporterEnum.HttpProtoOTLPExporter.value:
        from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter

        return OTLPLogExporter(**exporter_params)
    elif exporter_type == LoggingExporterEnum.GrpcProtoOTLPExporter.value:
        from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter

        return OTLPLogExporter(**exporter_params)
    elif exporter_type == LoggingExporterEnum.ConsoleLogExporter.value:
        from opentelemetry.sdk._logs.export import ConsoleLogExporter

        return ConsoleLogExporter(**exporter_params)

    raise ValueError(f"Invalid exporter type: {exporter_type}")


def build_log_record_processor(exporter: LogExporter, processor_config: dict) -> LogRecordProcessor:
    """Build an OpenTelemetry log record processor.
    params:
        processor_config: dict: The processor configuration.

    Returns:
        LogRecordProcessor: The log record processor instance.
    """
    processor_type = processor_config.get(
        "type", LogRecordProcessorEnum.SimpleLogRecordProcessor.value
    )
    processor_params = processor_config.get("params", {})

    if processor_type == LogRecordProcessorEnum.BatchLogRecordProcessor.value:
        from opentelemetry.sdk._logs.export import BatchLogRecordProcessor

        return BatchLogRecordProcessor(exporter, **processor_params)
    elif processor_type == LogRecordProcessorEnum.SimpleLogRecordProcessor.value:
        from opentelemetry.sdk._logs.export import SimpleLogRecordProcessor

        return SimpleLogRecordProcessor(exporter)

    raise ValueError(f"Invalid log record processor type: {processor_type}")


def build_logger_provider(
    processor: LogRecordProcessor,
    resource_attributes: Optional[dict[str, str]] = None,
) -> APILoggerProvider:
    """Build an OpenTelemetry logger provider.
    params:
        processor: LogRecordProcessor: The log record processor instance.
        resource: Resource (optional): The resource instance.

    Returns:
        LoggerProvider: The logger provider instance.
    """
    from opentelemetry.sdk._logs import LoggerProvider

    resource = (
        Resource.create(attributes=resource_attributes)
        if resource_attributes
        else Resource.create({})
    )

    logger_provider = LoggerProvider(
        resource=resource,
        shutdown_on_exit=False,
    )
    logger_provider.add_log_record_processor(processor)
    return logger_provider


def build_logging_handler(
    logger_provider: LoggerProvider, handler_config: Optional[dict] = None
) -> LoggingHandler:
    """Build an OpenTelemetry logging handler.
    Params:
        logger_provider (LoggerProvider): The logger provider instance.
        handler_config (Optional[dict]): The handler configuration.

    Returns:
        LoggingHandler: The logging handler instance.
    """
    from opentelemetry.sdk._logs import LoggingHandler

    handler_config = handler_config or {}

    return LoggingHandler(logger_provider=logger_provider, **handler_config)


def build_noop_logger_provider() -> APILoggerProvider:
    """Build a no-op OpenTelemetry meter provider."""
    from opentelemetry._logs import NoOpLoggerProvider

    return NoOpLoggerProvider()
